# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module provides Python event data structures to represent
interactions with the Pyre daemon, as well as an async utility
cann connect to a socket and wait (used when initializing new
daemons).
"""

import dataclasses
import enum
import json
from pathlib import Path
from typing import IO, List, Optional, Union

from ..language_server import connections
from .commands import ExitCode


@dataclasses.dataclass
class SocketCreated:
    socket_path: Path


@dataclasses.dataclass
class ServerInitialized:
    pass


class ErrorKind(enum.Enum):
    WATCHMAN = "Watchman"
    BUCK_INTERNAL = "BuckInternal"
    BUCK_USER = "BuckUser"
    PYRE = "Pyre"
    UNKNOWN = "Unknown"

    def __str__(self) -> str:
        return self.value

    @staticmethod
    def from_string(input_string: str) -> "ErrorKind":
        for item in ErrorKind:
            if input_string == str(item):
                return item
        return ErrorKind.UNKNOWN

    def to_exit_code(kind) -> ExitCode:
        if kind == ErrorKind.WATCHMAN:
            return ExitCode.WATCHMAN_ERROR
        elif kind == ErrorKind.BUCK_INTERNAL:
            return ExitCode.BUCK_INTERNAL_ERROR
        elif kind == ErrorKind.BUCK_USER:
            return ExitCode.BUCK_USER_ERROR
        return ExitCode.FAILURE


@dataclasses.dataclass
class ServerException:
    message: str
    kind: ErrorKind = ErrorKind.UNKNOWN


Event = Union[SocketCreated, ServerInitialized, ServerException]


def create_from_string(input_string: str) -> Optional[Event]:
    try:
        input_json: List[str] = json.loads(input_string)
        if len(input_json) < 1:
            return None

        input_kind = input_json[0]
        if input_kind == "SocketCreated":
            if len(input_json) < 2:
                return None
            else:
                return SocketCreated(socket_path=Path(input_json[1]))
        elif input_kind == "ServerInitialized":
            return ServerInitialized()
        elif input_kind == "Exception":
            if len(input_json) < 2:
                return None
            if not isinstance(input_json[1], str):
                return None
            if (
                len(input_json) >= 3
                and isinstance(input_json[2], list)
                and len(input_json[2]) > 0
                and isinstance(input_json[2][0], str)
            ):
                return ServerException(
                    message=input_json[1], kind=ErrorKind.from_string(input_json[2][0])
                )
            return ServerException(message=input_json[1], kind=ErrorKind.UNKNOWN)
        else:
            return None
    except json.JSONDecodeError:
        return None


class EventParsingException(Exception):
    pass


class ServerStartException(Exception):
    kind: ErrorKind

    def __init__(self, exception_event: ServerException) -> None:
        super().__init__(exception_event.message)
        self.kind = exception_event.kind


def _parse_server_event(event_string: str) -> Event:
    event = create_from_string(event_string)
    if event is None:
        raise EventParsingException(
            f"Unrecognized status update from server: {event_string}"
        )
    elif isinstance(event, ServerException):
        raise ServerStartException(event)
    return event


class Waiter:
    wait_on_initialization: bool

    def __init__(self, wait_on_initialization: bool) -> None:
        self.wait_on_initialization = wait_on_initialization

    def wait_on(self, event_stream: IO[str]) -> None:
        """
        Read from the given input channel, expecting server events there.
        If `self.wait_on_initialization` is false, block until server socket
        creation and returns.
        Otherwise, block until server initialization has finished and returns.
        If data obtained from the input channel does not conform to the server
        event format, raise `EventParsingException`. If an error event is received,
        raise `ServerStartException`.
        """
        # The first event is expected to be socket creation
        initial_event = _parse_server_event(event_stream.readline().strip())
        if isinstance(initial_event, SocketCreated):
            if not self.wait_on_initialization:
                return

            # The second event is expected to be server initialization
            second_event = _parse_server_event(event_stream.readline().strip())
            if isinstance(second_event, ServerInitialized):
                return

            raise EventParsingException(
                f"Unexpected second server status update: {second_event}"
            )

        raise EventParsingException(
            f"Unexpected initial server status update: {initial_event}"
        )

    # This method does the same thing as `wait_on` except it operates on asyncio
    # streams rather than synchronoized streams.
    # NOTE: Any changes inside `wait_on` need to be applied here as well.
    async def async_wait_on(self, event_stream: connections.AsyncTextReader) -> None:
        """
        This method is the same as `wait_on`, except it operates on async input
        channels instead of synchronous ones.
        """
        initial_event = _parse_server_event((await event_stream.readline()).strip())
        if isinstance(initial_event, SocketCreated):
            if not self.wait_on_initialization:
                return

            second_event = _parse_server_event((await event_stream.readline()).strip())
            if isinstance(second_event, ServerInitialized):
                return

            raise EventParsingException(
                f"Unexpected second server status update: {second_event}"
            )

        raise EventParsingException(
            f"Unexpected initial server status update: {initial_event}"
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import logging
from collections import defaultdict
from pathlib import Path
from typing import DefaultDict, List

import libcst
from libcst import metadata

from ..language_server import protocol as lsp

LOG: logging.Logger = logging.getLogger(__name__)

BUILTINS_BLOCKLIST = [
    "builtins.True",
    "builtins.False",
]


class QualifiedNameWithPositionVisitor(libcst.CSTVisitor):
    """
    A simple visitor that finds all the fully qualified names within a module.

    During the single passthrough, this visitor stores three things:
    1. All symbol positions per line
    2. For a given position of a symbol, a list of all the associated qualified names
    3. For a given qualified name, a list of all its associated positions.

    There are three main reasons for this:
    1. Storing position per line helps avoid having to re-traverse the tree to find the exact node.
    2. There are extremely rare scenarios a symbol can have multiple qualified names, which we need to aggregate for.
    3. Keeping a list of ranges for each qualified name allows constant time lookup of all references.

    Given its currently primary usecase is symbol renaming, the stored information is shortlived.

    TODO:
    1. Handle Import nodes. T164821394
    """

    METADATA_DEPENDENCIES = [
        metadata.FullyQualifiedNameProvider,
        metadata.PositionProvider,
    ]

    def __init__(self, position: lsp.LspPosition) -> None:
        self.name_to_ranges: DefaultDict[str, List[lsp.LspRange]] = defaultdict(list)
        self.qualified_names: List[str] = []
        self.symbol_range: lsp.LspRange = lsp.LspRange(
            start=lsp.LspPosition(line=0, character=0),
            end=lsp.LspPosition(line=0, character=0),
        )
        self.position: lsp.LspPosition = position

    def visit_Name(  # noqa: B906 - these are leaf nodes, no need to have recursive visits
        self, node: libcst.Name
    ) -> None:
        if self.get_metadata(metadata.PositionProvider, node):
            cst_position = self.get_metadata(metadata.PositionProvider, node)
            line = cst_position.start.line - 1
            lsp_range = lsp.LspRange(
                start=lsp.LspPosition(line=line, character=cst_position.start.column),
                end=lsp.LspPosition(line=line, character=cst_position.end.column),
            )

            if (
                line == self.position.line
                and lsp_range.start.character <= self.position.character
                and self.position.character <= lsp_range.end.character
            ):
                self.qualified_names = [
                    qualified_name.name
                    for qualified_name in self.get_metadata(
                        metadata.FullyQualifiedNameProvider, node
                    )
                ]
                self.symbol_range = lsp_range
            for qualified_name in self.get_metadata(
                metadata.FullyQualifiedNameProvider, node
            ):
                self.name_to_ranges[qualified_name.name].append(lsp_range)

    def find_references(self, include_builtins: bool = False) -> List[lsp.LspRange]:
        results = []
        for qualified_name in self.qualified_names:
            if include_builtins or qualified_name not in BUILTINS_BLOCKLIST:
                results.extend(self.name_to_ranges[qualified_name])
        return results

    def find_symbol_range(self) -> lsp.LspRange:
        return self.symbol_range


def generate_qualified_name_with_position_visitor(
    path: Path,
    global_root: Path,
    code: str,
    position: lsp.PyrePosition,
) -> QualifiedNameWithPositionVisitor:
    parsed = libcst.parse_module(code)
    fully_qualified_name = metadata.FullyQualifiedNameProvider.gen_cache(
        global_root, [str(path)]
    )
    wrapper = libcst.MetadataWrapper(
        parsed,
        cache={metadata.FullyQualifiedNameProvider: fully_qualified_name[str(path)]},
    )
    wrapper.resolve(metadata.FullyQualifiedNameProvider)
    wrapper.resolve(metadata.PositionProvider)
    visitor = QualifiedNameWithPositionVisitor(position.to_lsp_position())
    wrapper.visit(visitor)
    return visitor


def find_references(
    path: Path,
    global_root: Path,
    code: str,
    position: lsp.PyrePosition,
) -> List[lsp.LspLocation]:
    visitor = generate_qualified_name_with_position_visitor(
        path, global_root, code, position
    )
    references = visitor.find_references()
    return [
        lsp.LspLocation(
            uri=path.as_uri(),
            range=reference,
        )
        for reference in references
    ]


def find_symbol_range(
    path: Path,
    global_root: Path,
    code: str,
    position: lsp.PyrePosition,
) -> lsp.LspRange:
    visitor = generate_qualified_name_with_position_visitor(
        path, global_root, code, position
    )
    return visitor.find_symbol_range()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module provides shared logic (Persistent, CodeNav and Bento) for sending
status messages to the LSP client.
"""

import logging
from typing import Optional

from .. import json_rpc
from ..language_server import connections, features, protocol as lsp
from . import server_state

LOG: logging.Logger = logging.getLogger(__name__)


def _client_has_status_bar_support(
    client_capabilities: lsp.ClientCapabilities,
) -> bool:
    window_capabilities = client_capabilities.window
    if window_capabilities is not None:
        return window_capabilities.status is not None
    else:
        return False


async def _write_notification(
    output_channel: connections.AsyncTextWriter,
    message: str,
    short_message: Optional[str] = None,
    level: lsp.MessageType = lsp.MessageType.INFO,
) -> None:
    await lsp.write_json_rpc(
        output_channel,
        json_rpc.Request(
            method="window/showMessage",
            parameters=json_rpc.ByNameParameters(
                {
                    "type": int(level),
                    "message": (
                        message
                        if short_message is None
                        else f"{short_message}: {message}"
                    ),
                }
            ),
        ),
    )


async def _write_status(
    output_channel: connections.AsyncTextWriter,
    message: str,
    short_message: Optional[str] = None,
    level: lsp.MessageType = lsp.MessageType.INFO,
) -> None:
    await lsp.write_json_rpc(
        output_channel,
        json_rpc.Request(
            id=0,  # the value doesn't matter but the existence does
            method="window/showStatus",
            parameters=json_rpc.ByNameParameters(
                {
                    "type": int(level),
                    "message": message,
                    **(
                        {} if short_message is None else {"shortMessage": short_message}
                    ),
                }
            ),
        ),
    )


class ClientStatusMessageHandler:
    def __init__(
        self,
        client_output_channel: connections.AsyncTextWriter,
        server_state: server_state.ServerState,
    ) -> None:
        self.client_output_channel = client_output_channel
        self.server_state = server_state

    def get_status_updates_availability(self) -> features.StatusUpdatesAvailability:
        return self.server_state.server_options.language_server_features.status_updates

    async def show_notification_message_to_client(
        self,
        message: str,
        level: lsp.MessageType = lsp.MessageType.INFO,
    ) -> None:
        await _write_notification(self.client_output_channel, message, level=level)

    async def show_status_message_to_client(
        self,
        message: str,
        short_message: Optional[str] = None,
        level: lsp.MessageType = lsp.MessageType.INFO,
        fallback_to_notification: bool = False,
    ) -> None:
        if self.get_status_updates_availability().is_disabled():
            LOG.error("Status updates are disabled, skipping status message.")
            return
        if _client_has_status_bar_support(self.server_state.client_capabilities):
            await _write_status(
                self.client_output_channel, message, short_message, level
            )
        elif fallback_to_notification:
            await _write_notification(
                self.client_output_channel, message, short_message, level
            )

    def log(
        self,
        message: str,
        short_message: Optional[str] = None,
        level: lsp.MessageType = lsp.MessageType.INFO,
    ) -> None:
        log_message = (
            message if short_message is None else f"[{short_message}] {message}"
        )
        if level == lsp.MessageType.ERROR:
            LOG.error(log_message)
        elif level == lsp.MessageType.WARNING:
            LOG.warning(log_message)
        elif level == lsp.MessageType.INFO:
            LOG.info(log_message)
        else:
            LOG.debug(log_message)

    async def log_and_show_status_message_to_client(
        self,
        message: str,
        short_message: Optional[str] = None,
        level: lsp.MessageType = lsp.MessageType.INFO,
        fallback_to_notification: bool = False,
    ) -> None:
        self.log(message, short_message, level)
        await self.show_status_message_to_client(
            message, short_message, level, fallback_to_notification
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
The daemon query module contains APIs for sending requests to the Pyre daemon.
The responsibility of this layer is to serialize queries and send them over the connection,
but this module is not responsible for constructing the query/parsing the response
(which is left to the higher layers). It is also not responsible for the low-level details
of how the message is sent over the connection.

Note that this module does not handle requests of every format (such as overlay updates
nor incremental updates), it only handles queries.
"""

from __future__ import annotations

import json
import logging
from pathlib import Path
from typing import List, Optional, Type, TypeVar, Union

import dataclasses_json

from .. import dataclasses_json_extensions as json_mixins, error
from ..language_server import daemon_connection
from . import daemon_querier, incremental
from .query_response import InvalidQueryResponse, Response

QueryResponseType = TypeVar(
    "QueryResponseType", bound=json_mixins.CamlCaseAndExcludeJsonMixin
)

LOG: logging.Logger = logging.getLogger(__name__)


def execute_query(socket_path: Path, query_text: str) -> Response:
    raw_request = json.dumps(["Query", query_text])
    raw_response = daemon_connection.send_raw_request(socket_path, raw_request)
    return Response.parse(raw_response)


async def attempt_async_query(
    socket_path: Path,
    query_text: str,
    overlay_id: Optional[str] = None,
) -> Union[Response, daemon_querier.DaemonQueryFailure]:
    response_text = await daemon_connection.attempt_send_async_raw_request(
        socket_path=socket_path,
        request=json.dumps(
            # TODO:T126924773 send a regular Query when overlay id is None
            [
                "QueryWithOverlay",
                {"query_text": query_text, "overlay_id": overlay_id},
            ]
        ),
    )
    if isinstance(response_text, daemon_connection.DaemonConnectionFailure):
        return daemon_querier.DaemonQueryFailure(
            error_message=f"In attempt async query with response_text, got DaemonConnectionFailure exception: ({response_text.error_message})",
            error_source=response_text.error_source,
        )
    try:
        return Response.parse(response_text)
    except InvalidQueryResponse as exception:
        return daemon_querier.DaemonQueryFailure(
            f"In attempt async query with response_text, got InvalidQueryResponse exception: ({exception})"
        )


async def attempt_typed_async_query(
    response_type: Type[QueryResponseType],
    socket_path: Path,
    query_text: str,
    overlay_id: Optional[str] = None,
) -> Union[QueryResponseType | daemon_querier.DaemonQueryFailure]:
    try:
        response = await attempt_async_query(
            socket_path,
            query_text,
            overlay_id,
        )
        if isinstance(response, daemon_querier.DaemonQueryFailure):
            return response
        else:
            if not isinstance(response.payload, dict):
                raise ValueError(
                    f"Expected a dict, got {response.payload!r} as response"
                )
            if "error" in response.payload:
                return daemon_querier.DaemonQueryFailure(
                    f"Daemon query returned error: {response.payload} for query: {query_text}"
                )
            return response_type.from_dict(response.payload)
    except (
        KeyError,
        ValueError,
        dataclasses_json.mm.ValidationError,
    ) as exception:
        return daemon_querier.DaemonQueryFailure(
            f"When interpretting response type: {response_type.__name__} got: {type(exception).__name__}({exception})"
        )


async def attempt_async_overlay_type_errors(
    socket_path: Path,
    source_code_path: Path,
    overlay_id: str,
) -> Union[List[error.Error], daemon_querier.DaemonQueryFailure]:
    """
    In order to type check unsaved changes, we need to get type errors for one
    specific module in an overlay.

    We cannot use `attempt_async_query` for this because the output types are
    not quite compatible, but otherwise the logic works the same.
    """
    response_text = await daemon_connection.attempt_send_async_raw_request(
        socket_path=socket_path,
        request=json.dumps(
            [
                "GetOverlayTypeErrors",
                {"path": str(source_code_path), "overlay_id": overlay_id},
            ]
        ),
    )
    if isinstance(response_text, daemon_connection.DaemonConnectionFailure):
        return daemon_querier.DaemonQueryFailure(
            error_message=f"In attempt async query with response_text, got DaemonConnectionFailure exception: ({response_text.error_message})",
            error_source=response_text.error_source,
        )
    try:
        return incremental.parse_type_error_response(response_text).errors
    except incremental.InvalidServerResponse as exception:
        return daemon_querier.DaemonQueryFailure(
            f"In attempt async query with response_text, parsing led to uncaught error: ({exception})"
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
Entrypoint for the `pyre statistics` command.

This command allows users to get some simple information about the number of
code objects (e.g. returns, parameters), how many of those code objects have
annotations, and some other information such as the number of strict files
and pyre-fixme and pyre-ignore directives.

We intend to replace this command with the newer `pyre report`, which uses
a more ganular and tabular data format that will work better for downstream
purproses.

The `pyre statistics` command is in maintenance mode while we build out `pyre
report` and migrate use cases.
"""

import dataclasses
import json
import logging
import re
from pathlib import Path
from typing import Dict, Iterable, List, Mapping, Optional

import libcst
from libcst.metadata import CodePosition, CodeRange

from .. import command_arguments, coverage_data, frontend_configuration, log
from . import commands

LOG: logging.Logger = logging.getLogger(__name__)


def location_to_code_range(location: coverage_data.Location) -> CodeRange:
    """
    Convert a location back to a libcst.CodeRange.

    This is only needed for backward-compatibility of statistics json ouptut.
    """
    return CodeRange(
        start=CodePosition(
            line=location.start_line,
            column=location.start_column,
        ),
        end=CodePosition(
            line=location.end_line,
            column=location.end_column,
        ),
    )


@dataclasses.dataclass(frozen=True)
class ModuleSuppressionData:
    code: Dict[coverage_data.ErrorCode, List[coverage_data.LineNumber]]
    no_code: List[coverage_data.LineNumber]


class SuppressionCountCollector(coverage_data.VisitorWithPositionData):
    def __init__(self, regex: str) -> None:
        self.no_code: List[int] = []
        self.codes: Dict[int, List[int]] = {}
        self.regex: re.Pattern[str] = re.compile(regex)

    def error_codes(self, line: str) -> Optional[List[coverage_data.ErrorCode]]:
        match = self.regex.match(line)
        if match is None:
            # No suppression on line
            return None
        code_group = match.group(1)
        if code_group is None:
            # Code-less error suppression
            return []
        code_strings = code_group.strip("[] ").split(",")
        try:
            codes = [int(code) for code in code_strings]
            return codes
        except ValueError:
            LOG.warning("Invalid error suppression code: %s", line)
            return []

    def visit_Comment(self, node: libcst.Comment) -> None:
        error_codes = self.error_codes(node.value)
        if error_codes is None:
            return
        suppression_line = self.location(node).start_line
        if len(error_codes) == 0:
            self.no_code.append(suppression_line)
            return
        for code in error_codes:
            if code in self.codes:
                self.codes[code].append(suppression_line)
            else:
                self.codes[code] = [suppression_line]

    def collect(
        self,
        module: libcst.MetadataWrapper,
    ) -> ModuleSuppressionData:
        module.visit(self)
        return ModuleSuppressionData(code=self.codes, no_code=self.no_code)


class FixmeCountCollector(SuppressionCountCollector):
    def __init__(self) -> None:
        super().__init__(r".*# *pyre-fixme(\[(\d* *,? *)*\])?")


class IgnoreCountCollector(SuppressionCountCollector):
    def __init__(self) -> None:
        super().__init__(r".*# *pyre-ignore(\[(\d* *,? *)*\])?")


class TypeIgnoreCountCollector(SuppressionCountCollector):
    def __init__(self) -> None:
        super().__init__(r".*# *type: ignore")


@dataclasses.dataclass(frozen=True)
class ModuleAnnotationData:
    line_count: int
    total_functions: List[CodeRange]
    partially_annotated_functions: List[CodeRange]
    fully_annotated_functions: List[CodeRange]
    total_parameters: List[CodeRange]
    annotated_parameters: List[CodeRange]
    total_returns: List[CodeRange]
    annotated_returns: List[CodeRange]
    total_globals: List[CodeRange]
    annotated_globals: List[CodeRange]
    total_attributes: List[CodeRange]
    annotated_attributes: List[CodeRange]
    explicit_any_annotations: List[CodeRange]

    def to_count_dict(self) -> Dict[str, int]:
        return {
            "return_count": len(self.total_returns),
            "annotated_return_count": len(self.annotated_returns),
            "globals_count": len(self.total_globals),
            "annotated_globals_count": len(self.annotated_globals),
            "parameter_count": len(self.total_parameters),
            "annotated_parameter_count": len(self.annotated_parameters),
            "attribute_count": len(self.total_attributes),
            "annotated_attribute_count": len(self.annotated_attributes),
            "function_count": len(self.total_functions),
            "partially_annotated_function_count": len(
                self.partially_annotated_functions
            ),
            "fully_annotated_function_count": len(self.fully_annotated_functions),
            "line_count": self.line_count,
            "explicit_any_annotations": len(self.explicit_any_annotations),
        }


class AnnotationCountCollector(coverage_data.AnnotationCollector):
    def collect(
        self,
        module: libcst.MetadataWrapper,
    ) -> ModuleAnnotationData:
        module.visit(self)
        return ModuleAnnotationData(
            line_count=self.line_count,
            total_functions=[
                location_to_code_range(function.location) for function in self.functions
            ],
            partially_annotated_functions=[
                location_to_code_range(f.location)
                for f in self.functions
                if f.is_partially_annotated
            ],
            fully_annotated_functions=[
                location_to_code_range(f.location)
                for f in self.functions
                if f.is_fully_annotated
            ],
            total_parameters=[
                location_to_code_range(p.location) for p in list(self.parameters())
            ],
            annotated_parameters=[
                location_to_code_range(p.location)
                for p in self.parameters()
                if p.is_annotated
            ],
            total_returns=[location_to_code_range(r.location) for r in self.returns()],
            annotated_returns=[
                location_to_code_range(r.location)
                for r in self.returns()
                if r.is_annotated
            ],
            total_globals=[location_to_code_range(g.location) for g in self.globals],
            annotated_globals=[
                location_to_code_range(g.location)
                for g in self.globals
                if g.is_annotated
            ],
            total_attributes=[
                location_to_code_range(a.location) for a in self.attributes
            ],
            annotated_attributes=[
                location_to_code_range(a.location)
                for a in self.attributes
                if a.is_annotated
            ],
            explicit_any_annotations=[
                location_to_code_range(a.location)
                for a in self.parameters()
                if a.contains_explicit_any
            ]
            + [
                location_to_code_range(a.location)
                for a in self.returns()
                if a.contains_explicit_any
            ]
            + [
                location_to_code_range(a.location)
                for a in self.attributes
                if a.contains_explicit_any
            ],
        )


@dataclasses.dataclass(frozen=True)
class StatisticsData:
    annotations: ModuleAnnotationData
    fixmes: ModuleSuppressionData
    ignores: ModuleSuppressionData
    strict: coverage_data.ModuleModeInfo


def collect_statistics(
    sources: Iterable[Path], strict_default: bool
) -> Dict[str, StatisticsData]:
    data: Dict[str, StatisticsData] = {}
    for path in sources:
        module = coverage_data.module_from_path(path)
        if module is None:
            continue
        try:
            annotations = AnnotationCountCollector().collect(module)
            fixmes = FixmeCountCollector().collect(module)
            ignores = IgnoreCountCollector().collect(module)
            modes = coverage_data.collect_mode(module, strict_default, path)
            statistics_data = StatisticsData(
                annotations,
                fixmes,
                ignores,
                modes,
            )
            data[str(path)] = statistics_data
        except RecursionError:
            LOG.warning(f"LibCST encountered recursion error in `{path}`")
    return data


def collect_all_statistics(
    configuration: frontend_configuration.Base,
    statistics_arguments: command_arguments.StatisticsArguments,
) -> Dict[str, StatisticsData]:
    if statistics_arguments.paths is None:
        paths = [configuration.get_local_root() or configuration.get_global_root()]
    else:
        paths = statistics_arguments.paths
    return collect_statistics(
        coverage_data.find_module_paths(
            paths=paths,
            excludes=configuration.get_excludes(),
        ),
        strict_default=configuration.is_strict(),
    )


@dataclasses.dataclass(frozen=True)
class AggregatedStatisticsData:
    annotations: Dict[str, int] = dataclasses.field(default_factory=dict)
    fixmes: int = 0
    ignores: int = 0
    strict: int = 0
    unsafe: int = 0


def aggregate_statistics(
    data: Mapping[str, StatisticsData],
) -> AggregatedStatisticsData:
    aggregate_annotations = {
        "return_count": 0,
        "annotated_return_count": 0,
        "globals_count": 0,
        "annotated_globals_count": 0,
        "parameter_count": 0,
        "annotated_parameter_count": 0,
        "attribute_count": 0,
        "annotated_attribute_count": 0,
        "function_count": 0,
        "partially_annotated_function_count": 0,
        "fully_annotated_function_count": 0,
        "line_count": 0,
    }

    for statistics_data in data.values():
        annotation_counts = statistics_data.annotations.to_count_dict()
        for key in aggregate_annotations.keys():
            aggregate_annotations[key] += annotation_counts[key]

    return AggregatedStatisticsData(
        annotations=aggregate_annotations,
        fixmes=sum(
            len(fixmes.no_code) + len(fixmes.code)
            for fixmes in [statistics_data.fixmes for statistics_data in data.values()]
        ),
        ignores=sum(
            len(ignores.no_code) + len(ignores.code)
            for ignores in [
                statistics_data.ignores for statistics_data in data.values()
            ]
        ),
        strict=sum(
            1 if strictness.mode == coverage_data.ModuleMode.STRICT else 0
            for strictness in [
                statistics_data.strict for statistics_data in data.values()
            ]
        ),
        unsafe=sum(
            1 if strictness.mode == coverage_data.ModuleMode.UNSAFE else 0
            for strictness in [
                statistics_data.strict for statistics_data in data.values()
            ]
        ),
    )


def get_overall_annotation_percentage(data: AggregatedStatisticsData) -> float:
    total_annotation_slots = (
        data.annotations["return_count"]
        + data.annotations["parameter_count"]
        + data.annotations["attribute_count"]
        + data.annotations["globals_count"]
    )
    filled_annotation_slots = (
        data.annotations["annotated_return_count"]
        + data.annotations["annotated_parameter_count"]
        + data.annotations["annotated_attribute_count"]
        + data.annotations["annotated_globals_count"]
    )
    return filled_annotation_slots * 100.0 / total_annotation_slots


def get_summary(aggregated_data: AggregatedStatisticsData) -> str:
    annotation_rate = round(get_overall_annotation_percentage(aggregated_data), 2)
    total_suppressions = aggregated_data.fixmes + aggregated_data.ignores
    total_modules = aggregated_data.strict + aggregated_data.unsafe
    if total_modules > 0:
        strict_module_rate = round(aggregated_data.strict * 100.0 / total_modules, 2)
        unsafe_module_rate = round(aggregated_data.unsafe * 100.0 / total_modules, 2)
    else:
        strict_module_rate = 100.00
        unsafe_module_rate = 0.00
    total_function_count = aggregated_data.annotations["function_count"]
    if total_function_count > 0:
        partially_annotated_function_rate = round(
            aggregated_data.annotations["partially_annotated_function_count"]
            * 100.0
            / total_function_count,
            2,
        )
        fully_annotated_function_rate = round(
            aggregated_data.annotations["fully_annotated_function_count"]
            * 100.0
            / total_function_count,
            2,
        )
    else:
        partially_annotated_function_rate = 0.00
        fully_annotated_function_rate = 100.00
    return (
        f"Coverage summary:\nOverall annotation rate is {annotation_rate}%\n"
        + f"There are {total_suppressions} total error suppressions inline ({aggregated_data.fixmes} fixmes and {aggregated_data.ignores} ignores)\n"
        + f"Of {total_modules} modules, {strict_module_rate}% are strict and {unsafe_module_rate}% are unsafe\n"
        + f"Of {total_function_count} functions, {fully_annotated_function_rate}% are fully annotated and {partially_annotated_function_rate}% are partially annotated\n"
    )


def print_text_summary(data: Mapping[str, StatisticsData]) -> None:
    aggregated_data = aggregate_statistics(data)
    LOG.warning(get_summary(aggregated_data))


def print_json_summary(data: Mapping[str, StatisticsData]) -> None:
    aggregated_data = aggregate_statistics(data)
    log.stdout.write(json.dumps(dataclasses.asdict(aggregated_data), indent=4))


def process_json_statistics(
    data: Mapping[str, StatisticsData],
) -> Dict[str, Dict[str, Dict[str, int]]]:
    path_to_dictionary_statistics = {
        path: dataclasses.asdict(statistics_data)
        for (path, statistics_data) in data.items()
    }
    log.stdout.write(json.dumps(path_to_dictionary_statistics))
    return path_to_dictionary_statistics


def run_statistics(
    configuration: frontend_configuration.Base,
    statistics_arguments: command_arguments.StatisticsArguments,
) -> commands.ExitCode:
    data = collect_all_statistics(configuration, statistics_arguments)
    if statistics_arguments.print_summary:
        print_text_summary(data)
        return commands.ExitCode.SUCCESS

    if statistics_arguments.aggregate:
        print_json_summary(data)
        return commands.ExitCode.SUCCESS

    process_json_statistics(data)
    return commands.ExitCode.SUCCESS


def run(
    configuration: frontend_configuration.Base,
    statistics_arguments: command_arguments.StatisticsArguments,
) -> commands.ExitCode:
    LOG.info("Collecting statistics...")
    return run_statistics(configuration, statistics_arguments)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


"""
This is an implementation of Pysa's language server. It is a refactored
version of the server run by `pyre persistent`.
"""

import asyncio
import json
import logging
from typing import Union

from .. import (
    command_arguments,
    configuration as configuration_module,
    frontend_configuration,
    json_rpc,
    log,
    log_lsp_event,
)
from ..language_server import connections, protocol as lsp
from . import commands, start
from .initialization import (
    InitializationExit,
    InitializationFailure,
    InitializationSuccess,
)
from .persistent import process_initialize_request
from .pyre_language_server import _wait_for_exit, read_lsp_request

LOG: logging.Logger = logging.getLogger(__name__)


async def try_initialize(
    input_channel: connections.AsyncTextReader,
    output_channel: connections.AsyncTextWriter,
) -> Union[InitializationSuccess, InitializationFailure, InitializationExit]:
    """
    Read an LSP message from the input channel and try to initialize an LSP
    server. Also write to the output channel with proper response if the input
    message is a request instead of a notification.

    The function can return one of three possibilities:
    - If the initialization succeeds, return `InitializationSuccess`.
    - If the initialization fails, return `InitializationFailure`. There could
      be many reasons for the failure: The incoming LSP message may not be an
      initiailization request. The incoming LSP request may be malformed. Or the
      client may not complete the handshake by sending back an `initialized` request.
    - If an exit notification is received, return `InitializationExit`. The LSP
      spec allows exiting a server without a preceding initialize request.
    """
    request = None
    try:
        request = await lsp.read_json_rpc(input_channel)
        LOG.debug(f"Received pre-initialization LSP request: {request}")

        request_id = request.id
        if request_id is None:
            return (
                InitializationExit()
                if request.method == "exit"
                else InitializationFailure()
            )
        if request.method != "initialize":
            raise lsp.ServerNotInitializedError("An initialize request is needed.")
        request_parameters = request.parameters
        if request_parameters is None:
            raise lsp.ServerNotInitializedError(
                "Missing parameters for initialize request."
            )

        initialize_parameters = lsp.InitializeParameters.from_json_rpc_parameters(
            request_parameters
        )
        result = process_initialize_request(initialize_parameters)
        await lsp.write_json_rpc(
            output_channel,
            json_rpc.SuccessResponse(id=request_id, result=result.to_dict()),
        )

        initialized_notification = await lsp.read_json_rpc(input_channel)
        if initialized_notification.method == "shutdown":
            await _wait_for_exit(input_channel, output_channel)
            return InitializationExit()
        elif initialized_notification.method != "initialized":
            actual_message = json.dumps(initialized_notification.json())
            raise lsp.ServerNotInitializedError(
                "Failed to receive an `initialized` request from client. "
                + f"Got {log.truncate(actual_message, 100)}"
            )

        return InitializationSuccess(
            client_capabilities=initialize_parameters.capabilities,
            client_info=initialize_parameters.client_info,
            initialization_options=initialize_parameters.initialization_options,
        )
    except json_rpc.JSONRPCException as json_rpc_error:
        await lsp.write_json_rpc(
            output_channel,
            json_rpc.ErrorResponse(
                id=request.id if request is not None else None,
                code=json_rpc_error.error_code(),
                message=str(json_rpc_error),
                data={"retry": False},
            ),
        )
        return InitializationFailure(exception=json_rpc_error)


class PysaServer:
    # I/O Channels
    input_channel: connections.AsyncTextReader
    output_channel: connections.AsyncTextWriter

    # Immutable States
    client_capabilities: lsp.ClientCapabilities

    def __init__(
        self,
        input_channel: connections.AsyncTextReader,
        output_channel: connections.AsyncTextWriter,
        client_capabilities: lsp.ClientCapabilities,
        pyre_arguments: start.Arguments,
        binary_location: str,
        project_identifier: str,
    ) -> None:
        self.input_channel = input_channel
        self.output_channel = output_channel
        self.client_capabilities = client_capabilities
        self.pyre_arguments = pyre_arguments
        self.binary_location = binary_location
        self.project_identifier = project_identifier

    async def show_message_to_client(
        self, message: str, level: lsp.MessageType = lsp.MessageType.INFO
    ) -> None:
        await lsp.write_json_rpc(
            self.output_channel,
            json_rpc.Request(
                method="window/showMessage",
                parameters=json_rpc.ByNameParameters(
                    {"type": int(level), "message": message}
                ),
            ),
        )

    async def log_and_show_message_to_client(
        self, message: str, level: lsp.MessageType = lsp.MessageType.INFO
    ) -> None:
        if level == lsp.MessageType.ERROR:
            LOG.error(message)
        elif level == lsp.MessageType.WARNING:
            LOG.warning(message)
        elif level == lsp.MessageType.INFO:
            LOG.info(message)
        else:
            LOG.debug(message)
        await self.show_message_to_client(message, level)

    async def wait_for_exit(self) -> int:
        while True:
            request = await read_lsp_request(self.input_channel, self.output_channel)
            LOG.debug(f"Received post-shutdown request: {request}")

            if request.method == "exit":
                return 0
            else:
                raise json_rpc.InvalidRequestError("LSP server has been shut down")

    async def process_open_request(
        self, parameters: lsp.DidOpenTextDocumentParameters
    ) -> None:
        document_path = parameters.text_document.document_uri().to_file_path()
        if document_path is None:
            raise json_rpc.InvalidRequestError(
                f"Document URI is not a file: {parameters.text_document.uri}"
            )

    async def process_close_request(
        self, parameters: lsp.DidCloseTextDocumentParameters
    ) -> None:
        document_path = parameters.text_document.document_uri().to_file_path()
        if document_path is None:
            raise json_rpc.InvalidRequestError(
                f"Document URI is not a file: {parameters.text_document.uri}"
            )
        try:
            LOG.info(f"File closed: {document_path}")

        except KeyError:
            LOG.warning(f"Trying to close an un-opened file: {document_path}")

    async def process_did_save_request(
        self, parameters: lsp.DidSaveTextDocumentParameters
    ) -> None:
        document_path = parameters.text_document.document_uri().to_file_path()
        if document_path is None:
            raise json_rpc.InvalidRequestError(
                f"Document URI is not a file: {parameters.text_document.uri}"
            )

    async def run(self) -> int:
        while True:
            request = await read_lsp_request(self.input_channel, self.output_channel)

            try:
                if request.method == "exit":
                    return commands.ExitCode.FAILURE
                elif request.method == "shutdown":
                    await lsp.write_json_rpc(
                        self.output_channel,
                        json_rpc.SuccessResponse(id=request.id, result=None),
                    )
                    return await self.wait_for_exit()
                elif request.method == "textDocument/didOpen":
                    parameters = request.parameters
                    if parameters is None:
                        raise json_rpc.InvalidRequestError(
                            "Missing parameters for didOpen method"
                        )
                    await self.process_open_request(
                        lsp.DidOpenTextDocumentParameters.from_json_rpc_parameters(
                            parameters
                        )
                    )
                elif request.method == "textDocument/didClose":
                    parameters = request.parameters
                    if parameters is None:
                        raise json_rpc.InvalidRequestError(
                            "Missing parameters for didClose method"
                        )
                    await self.process_close_request(
                        lsp.DidCloseTextDocumentParameters.from_json_rpc_parameters(
                            parameters
                        )
                    )
                elif request.method == "textDocument/didSave":
                    parameters = request.parameters
                    if parameters is None:
                        raise json_rpc.InvalidRequestError(
                            "Missing parameters for didSave method"
                        )
                    await self.process_did_save_request(
                        lsp.DidSaveTextDocumentParameters.from_json_rpc_parameters(
                            parameters
                        )
                    )
                elif request.id is not None:
                    raise lsp.RequestCancelledError("Request not supported yet")
            except json_rpc.JSONRPCException as json_rpc_error:
                LOG.debug(
                    f"Exception occurred while processing request: {json_rpc_error}"
                )
                await lsp.write_json_rpc_ignore_connection_error(
                    self.output_channel,
                    json_rpc.ErrorResponse(
                        id=request.id,
                        activity_key=request.activity_key,
                        code=json_rpc_error.error_code(),
                        message=str(json_rpc_error),
                    ),
                )


async def run_persistent(
    binary_location: str, project_identifier: str, pysa_arguments: start.Arguments
) -> int:
    stdin, stdout = await connections.create_async_stdin_stdout()
    while True:
        initialize_result = await try_initialize(stdin, stdout)
        if isinstance(initialize_result, InitializationExit):
            LOG.info("Received exit request before initialization.")
            return 0
        elif isinstance(initialize_result, InitializationSuccess):
            LOG.info("Initialization successful.")
            client_info = initialize_result.client_info
            log_lsp_event.log(
                remote_logging=pysa_arguments.base_arguments.remote_logging,
                event=log_lsp_event.LSPEvent.INITIALIZED,
                normals=(
                    {}
                    if client_info is None
                    else {
                        "lsp client name": client_info.name,
                        "lsp client version": client_info.version,
                    }
                ),
            )

            client_capabilities = initialize_result.client_capabilities
            LOG.debug(f"Client capabilities: {client_capabilities}")
            server = PysaServer(
                input_channel=stdin,
                output_channel=stdout,
                client_capabilities=client_capabilities,
                binary_location=binary_location,
                project_identifier=project_identifier,
                pyre_arguments=pysa_arguments,
            )
            return await server.run()
        elif isinstance(initialize_result, InitializationFailure):
            exception = initialize_result.exception
            message = (
                str(exception) if exception is not None else "ignoring notification"
            )
            LOG.info(f"Initialization failed: {message}")
            # Loop until we get either InitializeExit or InitializeSuccess
        else:
            raise RuntimeError("Cannot determine the type of initialize_result")


def run(
    configuration: frontend_configuration.Base,
    start_arguments: command_arguments.StartArguments,
) -> int:
    start_command = configuration.get_server_start_command()
    if start_command is None:
        raise configuration_module.InvalidConfiguration(
            "Cannot locate a Pyre binary to run."
        )

    project_identifier = configuration.get_project_identifier()
    pyre_arguments = start.create_server_arguments(
        configuration,
        start_arguments,
        False,  # kill_buck_after_build
    )
    if pyre_arguments.watchman_root is None:
        raise commands.ClientException(
            (
                "Cannot locate a `watchman` root. Pyre's server will not function "
                + "properly."
            )
        )

    return asyncio.get_event_loop().run_until_complete(
        run_persistent(
            str(start_command.get_pyre_binary_location()),
            project_identifier,
            pyre_arguments,
        )
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module provides functionality related to Pyre's terminal logging,
including file tailing logic used to forward logs from the backend
and pretty-printing goodies such as the color-coding of interactive logs.
"""

import contextlib
import copy
import io
import logging
import logging.handlers
import re
import sys
import threading
import time
from pathlib import Path
from types import TracebackType
from typing import Dict, Generator, Iterable, Optional, Pattern, Sequence, TextIO

import click

PERFORMANCE: int = 15
PROMPT: int = 50
SUCCESS: int = 60


LOG: logging.Logger = logging.getLogger(__name__)
LOG_LEVELS: Dict[str, int] = {
    "CRITICAL": logging.CRITICAL,
    "ERROR": logging.ERROR,
    "WARNING": logging.WARNING,
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
}


stdout: io.StringIO = io.StringIO(newline="")
__handler: Optional["TimedStreamHandler"] = None


class Color:
    YELLOW: str = "\033[33m"
    RED: str = "\033[31m"
    GREEN: str = "\033[32m"


class Format:
    BOLD: str = "\033[1m"

    CLEAR_LINE: str = "\x1b[0G\x1b[K"
    CLEAR: str = "\033[0m"
    TRUNCATE_OVERFLOW: str = "\033[?7l"
    WRAP_OVERFLOW: str = "\033[?7h"
    NEWLINE: str = "\n"

    CURSOR_UP_LINE: str = "\x1b[1A"
    HIDE_CURSOR: str = "\x1b[?25l"
    SHOW_CURSOR: str = "\x1b[?25h"


class Character:
    LAMBDA: str = "ƛ"


class SectionFormatter(logging.Formatter):
    def __init__(self) -> None:
        super(SectionFormatter, self).__init__(
            "%(asctime)s [PID %(process)d] %(levelname)s %(message)s"
        )

    def format(self, record: logging.LogRecord) -> str:
        formatted = super(SectionFormatter, self).format(record)
        return re.sub(r"DEBUG \[(.*)\]", r"\1", formatted)


# pyre-fixme[24]: Generic type `logging.StreamHandler` expects 1 type parameter.
class TimedStreamHandler(logging.StreamHandler):
    THRESHOLD: float = 0.5
    LINE_BREAKING_LEVELS: Sequence[str] = ["ERROR", "WARNING", "SUCCESS"]

    _terminate: bool = False
    _last_update: float = 0.0

    def __init__(self) -> None:
        super(TimedStreamHandler, self).__init__()
        self.setFormatter(logging.Formatter("%(message)s"))
        self.terminator: str = ""
        self.setLevel(logging.INFO)

        self._record: Optional[logging.LogRecord] = None
        self._active_lines: int = 0

        # Preamble preparing terminal.
        click.echo(
            Format.NEWLINE
            + Format.CLEAR_LINE
            + Format.CURSOR_UP_LINE
            + Format.HIDE_CURSOR,
            file=sys.stderr,
            nl=False,
        )

        thread = threading.Thread(target=self._thread)
        thread.daemon = True
        thread.start()

    def clear_lines(self) -> str:
        if self._active_lines == 0:
            return ""
        return Format.CLEAR_LINE + "".join(
            [
                Format.CURSOR_UP_LINE + Format.CLEAR_LINE
                for n in range(self._active_lines - 1)
            ]
        )

    def emit(self, record: logging.LogRecord, age: Optional[float] = None) -> None:
        suffix = ""
        color: Optional[str] = None
        message = record.msg
        active_lines = message.count("\n") + 1
        truncate = Format.TRUNCATE_OVERFLOW
        if record.levelname in self.LINE_BREAKING_LEVELS:
            message += "\n"

        if record.levelname == "ERROR":
            color = "red"
            self._record = None
            active_lines = 0
            truncate = Format.WRAP_OVERFLOW
        elif record.levelname == "WARNING":
            color = "yellow"
            self._record = None
            active_lines = 0
            truncate = Format.WRAP_OVERFLOW
        elif record.levelname == "PROMPT":
            color = "yellow"
            self._record = None
            active_lines = 0
            truncate = Format.WRAP_OVERFLOW
        elif record.levelname == "SUCCESS":
            self._record = None
            active_lines = 0
            truncate = Format.WRAP_OVERFLOW
        elif age:
            if age > 10:
                color = "yellow"
            if age > 30:
                color = "red"
            suffix = click.style(f" [{age:.1f}s]", fg=color)
        else:
            self._record = record
            self._last_update = time.time()

        prompt = click.style(f"{Character.LAMBDA}", fg=color)
        new_message = f"{self.clear_lines()}{prompt} {truncate}{message}{suffix}"

        timed_record = copy.copy(record)
        timed_record.msg = (
            f"{click.unstyle(new_message)}\n"
            if click.utils.should_strip_ansi(stream=sys.stderr)
            else new_message
        )
        self._active_lines = active_lines
        super(TimedStreamHandler, self).emit(timed_record)

    def _thread(self) -> None:
        while not self._terminate:
            record = self._record
            if record:
                age = time.time() - self._last_update
                if age > self.THRESHOLD:
                    self.emit(record, age)
            time.sleep(0.1)

    def terminate(self) -> None:
        self._terminate = True

        if self._active_lines > 0:
            click.echo(self.clear_lines(), file=sys.stderr, nl=False)
            self._active_lines = 0

        # Reset terminal.
        click.echo(Format.WRAP_OVERFLOW + Format.SHOW_CURSOR, file=sys.stderr, nl=False)
        sys.stderr.flush()


def initialize(noninteractive: bool, *, logging_level: int = logging.DEBUG) -> None:
    global __handler

    if __handler:
        LOG.debug("Log handler already exists, skipping initialization.")
        return
    if noninteractive:
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(SectionFormatter())
        stream_handler.setLevel(logging_level)
        __handler = None
    else:
        stream_handler = TimedStreamHandler()
        __handler = stream_handler

    logging.addLevelName(PERFORMANCE, "PERFORMANCE")
    logging.addLevelName(PROMPT, "PROMPT")
    logging.addLevelName(SUCCESS, "SUCCESS")

    logging.basicConfig(level=logging_level, handlers=[stream_handler])


def enable_file_logging(log_file: Path) -> None:
    handler = logging.handlers.RotatingFileHandler(
        str(log_file),
        mode="a",
        # Keep at most 5 log files on disk
        backupCount=4,
        # Limit the size of each log file to 10MB
        maxBytes=10 * 1000 * 1000,
    )
    handler.setFormatter(SectionFormatter())
    handler.setLevel(logging.DEBUG)
    logger = logging.getLogger()
    logger.addHandler(handler)


def cleanup() -> None:
    global __handler
    handler = __handler
    if handler:
        handler.terminate()
        __handler = None

    output = stdout.getvalue()
    if output:
        click.echo(output, nl=False)
        if not output.endswith("\n"):
            click.echo()


@contextlib.contextmanager
def configured_logger(
    noninteractive: bool, *, logging_level: int = logging.DEBUG
) -> Generator[None, None, None]:
    try:
        initialize(noninteractive, logging_level=logging_level)
        yield
    finally:
        cleanup()


@contextlib.contextmanager
def file_tailer(file_path: Path) -> Generator[Iterable[str], None, None]:
    """
    This function yields a stream of string generated by following the last
    part of the given file. In other words, the returned stream behaves roughtly
    the same as `tail -F`: If the file being watched is left untouched, invoking
    `next` on the returned stream will block indefinitely. If the file being
    watched gets a line appended to the end of it, invoking `next` on the returned
    stream will return the appended line. Leaving the context manager will cause
    the returned stream to stop iteration next time `next` is invoked.

    This API is intended to be used along with `StreamLogger` to concurrently
    forward the content of a log file to the terminal in the background:

    ```
    with file_tailer(log_file) as log_stream:
        with StreamLogger(log_stream) as logger:
            # Main thread logic happens here
            ...
    logger.join()
    ```
    """
    stop_reader: bool = False

    def follow(file_path: Path) -> Iterable[str]:
        file: Optional[TextIO] = None
        sleep_interval_seconds = 0.01

        def safely_close_file() -> None:
            nonlocal file

            if file is None:
                return

            try:
                file.close()
            except Exception:
                pass

            file = None

        try:
            while not stop_reader:
                try:
                    if file is None:
                        file = open(file_path, "r")
                        file.seek(0, 2)
                    else:
                        line = file.readline()

                        if len(line) == 0:
                            time.sleep(sleep_interval_seconds)
                            continue

                        yield line
                except Exception:
                    # exception on open(), file.seek() or file.readline()
                    # We manually use `open()` and `close()` here instead of using
                    # `with open()`, since it would require using both an outer and
                    # inner `while` loop and would lead to code duplication.
                    safely_close_file()
                    time.sleep(sleep_interval_seconds)
        finally:
            safely_close_file()

    yield follow(file_path)
    stop_reader = True


class StreamLogger:
    _should_stop_reading_stream = False
    _current_section: Optional[str]

    _server_log_pattern: Pattern[str] = re.compile(
        r"\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2} (\w+)(.*)"
    )

    def __init__(self, stream: Iterable[str]) -> None:
        self._reader = threading.Thread(target=self._read_stream, args=(stream,))
        self._reader.daemon = True
        self._current_section = None

    def join(self) -> None:
        self._reader.join()

    def _log_server_stderr_message(self, server_message: str) -> None:
        line = server_message.rstrip()
        match = self._server_log_pattern.match(line)
        if match:
            section = match.groups()[0]
            message = match.groups()[1]
            self._current_section = section
        else:
            section = self._current_section
            message = line

        if section == "ERROR":
            LOG.error(message)
        elif section == "INFO":
            LOG.info(message)
        elif section == "DUMP":
            LOG.warning(message)
        elif section == "WARNING":
            LOG.warning(message)
        elif section == "PROGRESS":
            LOG.info(message)
        elif section == "PARSER":
            LOG.error(message)
        elif section is not None:
            LOG.debug("[%s] %s", section, message)
        else:
            LOG.debug(line)

    def _read_stream(self, stream: Iterable[str]) -> None:
        try:
            for line in stream:
                if self._should_stop_reading_stream:
                    return
                self._log_server_stderr_message(line)
        except Exception:
            pass

    def __enter__(self) -> "StreamLogger":
        self._should_stop_reading_stream = False
        self._reader.start()
        return self

    def __exit__(
        self,
        _type: Optional[BaseException],
        _value: Optional[BaseException],
        _traceback: Optional[TracebackType],
    ) -> None:
        self._should_stop_reading_stream = True


def get_yes_no_input(prompt: str) -> bool:
    choice = get_input(prompt, suffix=" [Y/n] ")
    return choice.lower() in ["", "y", "ye", "yes"]


def get_optional_input(prompt: str, default: str) -> str:
    result = get_input(prompt, suffix=f" (Default: `{default}`): ")
    if result == "":
        return default
    return result


def get_input(prompt: str, suffix: str = "") -> str:
    LOG.log(PROMPT, prompt + suffix)
    return input().strip()


def truncate(message: str, size: int) -> str:
    if len(message) <= size:
        return message

    return f"{message[:size]}..[truncated {len(message) - size} characters]"



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module provides the logic for `pyre start`, which starts up a Pyre
daemon.

In most cases daemons are started incrementally when starting a language
server or running incremental checks, but it can be useful to have a
dedicated start command, particularly to start a deamon that we can later
query.
"""

import contextlib
import dataclasses
import datetime
import enum
import logging
import os
import subprocess
import tempfile
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    TextIO,
    Tuple,
    Union,
)

from .. import (
    backend_arguments,
    command_arguments,
    configuration as configuration_module,
    daemon_socket,
    find_directories,
    frontend_configuration,
    identifiers,
    log,
)
from . import commands, server_event, stop

LOG: logging.Logger = logging.getLogger(__name__)

DAEMON_LOG_PATH_FORMAT: str = "server{flavor_suffix}.stderr.{time}"
DAEMON_LOG_TIME_FORMAT: str = "%Y_%m_%d_%H_%M_%S_%f"
DAEMON_CURRENT_LOG_PATH_FORMAT: str = "server{flavor_suffix}.stderr"


def deamon_current_log_path(
    log_directory: Path,
    flavor: identifiers.PyreFlavor,
) -> Path:
    return log_directory / DAEMON_CURRENT_LOG_PATH_FORMAT.format(
        flavor_suffix=flavor.path_suffix()
    )


def daemon_log_path(
    log_directory: Path,
    flavor: identifiers.PyreFlavor,
    now: datetime.datetime,
) -> Path:
    return log_directory / DAEMON_LOG_PATH_FORMAT.format(
        flavor_suffix=flavor.path_suffix(),
        time=now.strftime(DAEMON_LOG_TIME_FORMAT),
    )


def datetime_from_log_path(path: Path) -> Optional[datetime.datetime]:
    try:
        time_portion = path.name.split(".")[-1]
        return datetime.datetime.strptime(time_portion, DAEMON_LOG_TIME_FORMAT)
    except (IndexError, ValueError):
        return None


class MatchPolicy(enum.Enum):
    BASE_NAME = "base_name"
    EXTENSION = "extension"
    FULL_PATH = "full_path"

    def __str__(self) -> str:
        return self.value


@dataclasses.dataclass(frozen=True)
class CriticalFile:
    policy: MatchPolicy
    path: str

    def serialize(self) -> Dict[str, str]:
        return {str(self.policy): self.path}


@dataclasses.dataclass(frozen=True)
class LoadSavedStateFromFile:
    shared_memory_path: str
    changed_files_path: Optional[str] = None

    def serialize(self) -> Tuple[str, Dict[str, str]]:
        return (
            "load_from_file",
            {
                "shared_memory_path": self.shared_memory_path,
                **(
                    {}
                    if self.changed_files_path is None
                    else {"changed_files_path": self.changed_files_path}
                ),
            },
        )


@dataclasses.dataclass(frozen=True)
class LoadSavedStateFromProject:
    project_name: str
    project_metadata: Optional[str] = None

    def serialize(self) -> Tuple[str, Dict[str, str]]:
        return (
            "load_from_project",
            {
                "project_name": self.project_name,
                **(
                    {}
                    if self.project_metadata is None
                    else {"project_metadata": self.project_metadata}
                ),
            },
        )


@dataclasses.dataclass(frozen=True)
class StoreSavedStateToFile:
    shared_memory_path: str

    def serialize(self) -> Tuple[str, Dict[str, str]]:
        return (
            "save_to_file",
            {
                "shared_memory_path": self.shared_memory_path,
            },
        )


SavedStateAction = Union[
    LoadSavedStateFromFile, LoadSavedStateFromProject, StoreSavedStateToFile
]


@dataclasses.dataclass(frozen=True)
class Arguments:
    """
    Data structure for configuration options the backend server can recognize.
    Need to keep in sync with `source/command/serverCommand.ml`
    """

    base_arguments: backend_arguments.BaseArguments

    socket_path: Path
    strict: bool = False
    show_error_traces: bool = False
    additional_logging_sections: Sequence[str] = dataclasses.field(default_factory=list)
    watchman_root: Optional[Path] = None
    taint_models_path: Sequence[str] = dataclasses.field(default_factory=list)
    store_type_check_resolution: bool = False
    critical_files: Sequence[CriticalFile] = dataclasses.field(default_factory=list)
    saved_state_action: Optional[SavedStateAction] = None
    skip_initial_type_check: bool = False
    use_lazy_module_tracking: bool = False
    analyze_external_sources: bool = False

    def serialize(self) -> Dict[str, Any]:
        return {
            **self.base_arguments.serialize(),
            "strict": self.strict,
            "socket_path": str(self.socket_path),
            "show_error_traces": self.show_error_traces,
            "additional_logging_sections": self.additional_logging_sections,
            **(
                {}
                if self.watchman_root is None
                else {"watchman_root": str(self.watchman_root)}
            ),
            "taint_model_paths": self.taint_models_path,
            "store_type_check_resolution": self.store_type_check_resolution,
            "critical_files": [
                critical_file.serialize() for critical_file in self.critical_files
            ],
            **(
                {}
                if self.saved_state_action is None
                else {"saved_state_action": self.saved_state_action.serialize()}
            ),
            "skip_initial_type_check": self.skip_initial_type_check,
            "use_lazy_module_tracking": self.use_lazy_module_tracking,
            "analyze_external_sources": self.analyze_external_sources,
        }


def get_critical_files(
    configuration: frontend_configuration.Base,
    flavor: identifiers.PyreFlavor = identifiers.PyreFlavor.CLASSIC,
) -> List[CriticalFile]:
    def get_full_path(root: str, relative: str) -> str:
        full_path = (Path(root) / relative).resolve(strict=False)
        if not full_path.exists():
            LOG.warning(f"Critical file does not exist: {full_path}")
        return str(full_path)

    # TODO(T137504540) update critical files for overridden configs as well
    configuration_name = find_directories.JSON_CONFIGURATION_FILE
    local_root = configuration.get_local_root()
    return [
        CriticalFile(
            policy=MatchPolicy.FULL_PATH,
            path=get_full_path(
                root=str(configuration.get_global_root()),
                relative=configuration_name,
            ),
        ),
        *(
            []
            if local_root is None
            else [
                CriticalFile(
                    policy=MatchPolicy.FULL_PATH,
                    path=get_full_path(
                        root=str(local_root),
                        relative=find_directories.LOCAL_CONFIGURATION_FILE,
                    ),
                )
            ]
        ),
        *(
            # TODO(T92070475): This is a temporary hack until generated code can be
            # fully supported.
            []
            if configuration.get_buck_targets() is None
            else [CriticalFile(policy=MatchPolicy.EXTENSION, path="thrift")]
        ),
        *(
            [
                CriticalFile(
                    policy=MatchPolicy.FULL_PATH,
                    path=get_full_path(root=path, relative=""),
                )
                for path in configuration.get_other_critical_files()
            ]
        ),
    ]


def get_saved_state_action(
    start_arguments: command_arguments.StartArguments,
    configuration: frontend_configuration.Base,
    relative_local_root: Optional[str] = None,
) -> Optional[SavedStateAction]:
    saved_state_output_path = start_arguments.save_initial_state_to
    if saved_state_output_path is not None:
        return StoreSavedStateToFile(shared_memory_path=saved_state_output_path)

    saved_state_input_path = start_arguments.load_initial_state_from
    if saved_state_input_path is not None:
        return LoadSavedStateFromFile(
            shared_memory_path=saved_state_input_path,
            changed_files_path=start_arguments.changed_files_path,
        )

    saved_state_project = configuration.get_saved_state_project()
    if saved_state_project is not None:
        return LoadSavedStateFromProject(
            project_name=saved_state_project.name,
            project_metadata=saved_state_project.metadata,
        )

    return None


def create_server_arguments(
    configuration: frontend_configuration.Base,
    start_arguments: command_arguments.StartArguments,
    kill_buck_after_build: bool,
) -> Arguments:
    """
    Translate client configurations and command-line flags to server
    configurations.

    This API is not pure since it needs to access filesystem to filter out
    nonexistent directories. It is idempotent though, since it does not alter
    any filesystem state.
    """
    global_root = configuration.get_global_root()
    relative_local_root = configuration.get_relative_local_root()
    watchman_root = (
        None
        if start_arguments.no_watchman
        else backend_arguments.find_watchman_root(global_root)
    )

    source_paths = backend_arguments.get_source_path_for_server(
        configuration,
        start_arguments.flavor,
        kill_buck_after_build,
        start_arguments.number_of_buck_threads,
        watchman_root,
    )

    logging_sections = start_arguments.logging_sections
    additional_logging_sections = (
        [] if logging_sections is None else logging_sections.split(",")
    )
    if start_arguments.noninteractive:
        additional_logging_sections.append("-progress")
    # Server section is usually useful when Pyre server is involved
    additional_logging_sections.append("server")

    log_directory = configuration.get_log_directory()
    profiling_output = (
        backend_arguments.get_profiling_log_path(log_directory)
        if start_arguments.enable_profiling
        else None
    )
    memory_profiling_output = (
        backend_arguments.get_profiling_log_path(log_directory)
        if start_arguments.enable_memory_profiling
        else None
    )

    return Arguments(
        base_arguments=backend_arguments.BaseArguments(
            log_path=str(log_directory),
            global_root=str(global_root),
            checked_directory_allowlist=backend_arguments.get_checked_directory_allowlist(
                configuration, source_paths
            ),
            checked_directory_blocklist=(configuration.get_ignore_all_errors()),
            debug=start_arguments.debug,
            enable_readonly_analysis=configuration.get_enable_readonly_analysis(),
            enable_strict_override_check=configuration.get_enable_strict_override_check(),
            enable_strict_any_check=configuration.get_enable_strict_any_check(),
            enable_unawaited_awaitable_analysis=(
                configuration.get_enable_unawaited_awaitable_analysis()
            ),
            excludes=configuration.get_excludes(),
            extensions=configuration.get_valid_extension_suffixes(),
            include_suppressed_errors=configuration.get_include_suppressed_errors(),
            relative_local_root=relative_local_root,
            memory_profiling_output=memory_profiling_output,
            number_of_workers=configuration.get_number_of_workers(),
            parallel=not start_arguments.sequential,
            profiling_output=profiling_output,
            python_version=configuration.get_python_version(),
            system_platform=configuration.get_system_platform(),
            shared_memory=configuration.get_shared_memory(),
            remote_logging=backend_arguments.RemoteLogging.create(
                configuration.get_remote_logger(),
                start_arguments.get_log_identifier(),
            ),
            search_paths=configuration.get_existent_search_paths(),
            source_paths=source_paths,
        ),
        strict=configuration.is_strict(),
        show_error_traces=start_arguments.show_error_traces,
        additional_logging_sections=additional_logging_sections,
        watchman_root=watchman_root,
        taint_models_path=configuration.get_taint_models_path(),
        store_type_check_resolution=start_arguments.store_type_check_resolution,
        critical_files=get_critical_files(configuration, start_arguments.flavor),
        saved_state_action=(
            None
            if start_arguments.no_saved_state
            else get_saved_state_action(
                start_arguments, configuration, relative_local_root=relative_local_root
            )
        ),
        skip_initial_type_check=start_arguments.skip_initial_type_check,
        use_lazy_module_tracking=start_arguments.use_lazy_module_tracking,
        analyze_external_sources=start_arguments.analyze_external_sources,
        socket_path=daemon_socket.get_socket_path(
            configuration.get_project_identifier(),
            flavor=start_arguments.flavor,
        ),
    )


def _run_in_foreground(
    command: Sequence[str], environment: Mapping[str, str]
) -> commands.ExitCode:
    # In foreground mode, we shell out to the backend server and block on it.
    # Server stdout/stderr will be forwarded to the current terminal.
    return_code = 0
    try:
        LOG.warning("Starting server in the foreground...")
        # lint-ignore: NoUnsafeExecRule
        result = subprocess.run(
            command,
            env=environment,
            stdout=subprocess.DEVNULL,
            stderr=None,
            universal_newlines=True,
        )
        return_code = result.returncode
    except KeyboardInterrupt:
        # Backend server will exit cleanly when receiving SIGINT.
        pass

    if return_code == 0:
        return commands.ExitCode.SUCCESS
    else:
        LOG.error(f"Server exited with non-zero return code: {return_code}")
        return commands.ExitCode.FAILURE


@contextlib.contextmanager
def background_logging(log_file: Path) -> Iterator[None]:
    with log.file_tailer(log_file) as log_stream:
        with log.StreamLogger(log_stream) as logger:
            yield
    logger.join()


def _create_symbolic_link(source: Path, target: Path) -> None:
    try:
        source.unlink()
    except FileNotFoundError:
        pass
    source.symlink_to(target)


@contextlib.contextmanager
def background_server_log_file(
    log_directory: Path,
    flavor: identifiers.PyreFlavor,
) -> Iterator[TextIO]:
    new_server_log_directory = log_directory / flavor.server_log_subdirectory()
    new_server_log_directory.mkdir(parents=True, exist_ok=True)
    log_file_path = daemon_log_path(
        log_directory=new_server_log_directory,
        flavor=flavor,
        now=datetime.datetime.now(),
    )
    # lint-ignore: NoUnsafeFilesystemRule
    with open(str(log_file_path), "a") as log_file:
        yield log_file
    # Symlink the log file to a known location for subsequent `pyre incremental`
    # to find.
    _create_symbolic_link(
        deamon_current_log_path(
            log_directory=new_server_log_directory,
            flavor=flavor,
        ),
        log_file_path,
    )


def _run_in_background(
    command: Sequence[str],
    environment: Mapping[str, str],
    log_directory: Path,
    socket_path: Path,
    flavor: identifiers.PyreFlavor,
    event_waiter: server_event.Waiter,
) -> commands.ExitCode:
    # In background mode, we asynchronously start the server with `Popen` and
    # detach it from the current process immediately with `start_new_session`.
    # Do not call `wait()` on the Popen object to avoid blocking.
    # Server stderr will be forwarded to dedicated log files.
    # Server stdout will be used as additional communication channel for status
    # updates.
    with background_server_log_file(log_directory, flavor=flavor) as server_stderr:
        log_file = Path(server_stderr.name)
        # lint-ignore: NoUnsafeExecRule
        server_process = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=server_stderr,
            env=environment,
            start_new_session=True,
            universal_newlines=True,
        )

    LOG.info("Server is starting in the background.\n")
    server_stdout = server_process.stdout
    if server_stdout is None:
        raise RuntimeError("subprocess.Popen failed to set up a pipe for server stdout")

    try:
        # Block until an expected server event is obtained from stdout
        with background_logging(log_file):
            event_waiter.wait_on(server_stdout)
            server_stdout.close()
            return commands.ExitCode.SUCCESS
    except KeyboardInterrupt:
        LOG.info("SIGINT received. Terminating background server...")

        # If a background server has spawned and the user hits Ctrl-C, bring down
        # the spwaned server as well.
        server_stdout.close()
        server_process.terminate()
        server_process.wait()

        # Since we abruptly terminate the background server, it may not have the
        # chance to clean up the socket file properly. Make sure the file is
        # removed on our side.
        stop.remove_socket_if_exists(socket_path)

        raise commands.ClientException(
            "Interrupted by user. No server is spawned."
        ) from None


def run(
    configuration: frontend_configuration.Base,
    start_arguments: command_arguments.StartArguments,
) -> commands.ExitCode:
    server_start_command = configuration.get_server_start_command(
        download_if_needed=True
    )
    if server_start_command is None:
        raise configuration_module.InvalidConfiguration(
            "Cannot locate a Pyre binary to run."
        )
    LOG.info(
        f"Pyre binary is located at `{server_start_command.get_pyre_binary_location()}`"
    )

    log_directory = configuration.get_log_directory()
    server_arguments = create_server_arguments(
        configuration,
        start_arguments,
        False,  # kill_buck_after_build
    )
    if not start_arguments.no_watchman and server_arguments.watchman_root is None:
        LOG.warning(
            "Cannot find a watchman root. Incremental check will not be functional"
            " since no filesystem updates will be sent to the Pyre server."
        )

    LOG.info(f"Starting server at `{configuration.get_project_identifier()}`...")
    try:
        with backend_arguments.temporary_argument_file(
            server_arguments
        ) as argument_file_path:
            server_environment = {
                **os.environ,
                # This is to make sure that backend server shares the socket root
                # directory with the client.
                # TODO(T77556312): It might be cleaner to turn this into a
                # configuration option instead.
                "TMPDIR": tempfile.gettempdir(),
            }
            start_command = server_start_command.get_start_command(
                argument_file_path, start_arguments.flavor
            )
            if start_arguments.terminal:
                return _run_in_foreground(
                    start_command,
                    server_environment,
                )
            else:
                socket_path = daemon_socket.get_socket_path(
                    configuration.get_project_identifier(),
                    flavor=start_arguments.flavor,
                )
                return _run_in_background(
                    start_command,
                    server_environment,
                    log_directory,
                    socket_path,
                    flavor=start_arguments.flavor,
                    event_waiter=server_event.Waiter(
                        wait_on_initialization=start_arguments.wait_on_initialization
                    ),
                )
    except server_event.ServerStartException as error:
        raise commands.ClientException(
            f"{error}", exit_code=error.kind.to_exit_code()
        ) from error



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
TODO(T132414938) Add a module-level docstring
"""

from __future__ import annotations

import json
import logging
import re
import subprocess
from logging import Logger
from pathlib import Path
from typing import Any, Dict, Generator, List, Optional, Sequence, Set

from . import UserError
from .errors import Errors
from .filesystem import get_filesystem

LOG: Logger = logging.getLogger(__name__)


class Configuration:
    def __init__(
        self, path: Path, json_contents: Optional[Dict[str, Any]] = None
    ) -> None:
        if json_contents is None:
            with open(path, "r") as configuration_file:
                json_contents = json.load(configuration_file)
        self._path: Path = path
        if path.name == ".pyre_configuration.local":
            self.is_local: bool = True
        else:
            self.is_local: bool = False
        self.root: str = str(path.parent)
        self.original_contents: Dict[str, Any] = json_contents

        # Configuration fields
        self.strict: Optional[bool] = json_contents.get("strict")
        self.targets: Optional[List[str]] = json_contents.get("targets")
        self.source_directories: Optional[List[str]] = json_contents.get(
            "source_directories"
        )
        self.version: Optional[str] = json_contents.get("version")
        self.pysa_version: Optional[str] = json_contents.get("pysa_version")
        self.use_buck_builder: Optional[bool] = json_contents.get("use_buck_builder")
        self.use_buck_source_database: Optional[bool] = json_contents.get(
            "use_buck_source_database"
        )
        self.ignore_all_errors: Optional[List[str]] = json_contents.get(
            "ignore_all_errors"
        )
        self.exclude: Optional[List[str]] = json_contents.get("exclude")
        self.use_buck2: Optional[bool] = json_contents.get("use_buck2")

    def get_contents(self) -> Dict[str, Any]:
        """Assumption: The field names in this class match the key names in
        the configuration."""
        contents: Dict[str, Any] = self.original_contents

        def update_contents(key: str) -> None:
            attribute = getattr(self, key)
            if attribute is not None:
                contents[key] = attribute
            elif key in contents:
                del contents[key]

        update_contents("targets")
        update_contents("source_directories")
        update_contents("version")
        update_contents("pysa_version")
        update_contents("strict")
        update_contents("use_buck_builder")
        update_contents("use_buck_source_database")
        update_contents("use_buck2")
        return contents

    @staticmethod
    def find_parent_file(
        filename: str, directory: Optional[Path] = None
    ) -> Optional[Path]:
        directory = directory or Path.cwd()
        root = directory.root
        while directory != root:
            configuration_path = directory / filename
            if configuration_path.is_file():
                return configuration_path
            parent = directory.parent
            if directory == parent:
                return None
            directory = parent
        return None

    @staticmethod
    def find_project_configuration(directory: Optional[Path] = None) -> Path:
        path = Configuration.find_parent_file(".pyre_configuration", directory)
        if path is None:
            raise UserError("No root with a `.pyre_configuration` found.")
        return path

    @staticmethod
    def find_local_configuration(directory: Optional[Path] = None) -> Optional[Path]:
        return Configuration.find_parent_file(".pyre_configuration.local", directory)

    @staticmethod
    def gather_local_configuration_paths(directory: str) -> Sequence[Path]:
        return [
            Path(path)
            for path in get_filesystem().list(
                directory, patterns=[r"**\.pyre_configuration.local"]
            )
        ]

    @staticmethod
    def gather_local_configurations() -> List["Configuration"]:
        LOG.info("Finding configurations...")
        configuration_paths = Configuration.gather_local_configuration_paths(".")
        if not configuration_paths:
            LOG.info("No projects with local configurations found.")
            return []
        configurations = []
        for configuration_path in configuration_paths:
            with open(configuration_path) as configuration_file:
                try:
                    configuration = Configuration(
                        configuration_path, json.load(configuration_file)
                    )
                    configurations.append(configuration)
                except json.decoder.JSONDecodeError:
                    LOG.error(
                        "Configuration at `%s` is invalid, skipping.",
                        configuration_path,
                    )
        LOG.info(
            "Found %d local configuration%s.",
            len(configurations),
            "s" if len(configurations) != 1 else "",
        )
        return configurations

    def get_path(self) -> Path:
        return self._path

    def get_source_paths(self) -> Generator[Path, None, None]:
        # This is an approximation
        return Path(self.root).glob("**/*.py")

    def get_exclude_as_patterns(self) -> Set[re.Pattern[str]]:
        if self.exclude is not None:
            return {re.compile(pattern) for pattern in self.exclude}
        else:
            return set()

    def _relative_path_from_prefix(self, prefix: str) -> Path:
        root = Path(self.root)
        if not prefix.startswith("//"):
            return root / prefix

        # strip preceding '//'
        stripped_prefix = prefix[2:]

        if self.is_local:
            global_root = self.find_project_configuration(root).parent
            return global_root / stripped_prefix
        else:
            return root / stripped_prefix

    def get_resolved_ignore_path_prefixes(self) -> Set[Path]:
        if self.ignore_all_errors is not None:
            return {
                self._relative_path_from_prefix(prefix).resolve()
                for prefix in self.ignore_all_errors
            }
        else:
            return set()

    def get_directory(self) -> Path:
        return self._path.parent

    def write(self) -> None:
        with open(self._path, "w") as configuration_file:
            json.dump(self.get_contents(), configuration_file, sort_keys=True, indent=2)
            configuration_file.write("\n")

    def remove_version(self) -> None:
        if not self.version:
            LOG.info("Version not found in configuration.")
            return
        self.version = None

    def delete(self) -> None:
        self._path.unlink()

    def set_version(self, version: str) -> None:
        self.version = version

    def set_pysa_version(self, pysa_version: str) -> None:
        self.pysa_version = pysa_version

    def enable_source_database_buck_builder(self) -> None:
        self.use_buck_builder = True
        self.use_buck_source_database = True

    def set_use_buck1_if_possible(self) -> None:
        if self.use_buck2 is None:
            self.use_buck2 = False

    def add_strict(self) -> None:
        if self.strict:
            LOG.info("Configuration is already strict.")
            return
        self.strict = True

    # Pyre is now strict by default, so in most cases we want to
    # remove the strict flag from the config file.
    def use_strict_default(self) -> None:
        self.strict = None

    def add_targets(self, targets: List[str]) -> None:
        existing_targets = self.targets
        if existing_targets:
            existing_targets.extend(targets)
        else:
            self.targets = targets

    def has_single_wildcard_target(self) -> bool:
        return (
            self.targets is not None
            and len(self.targets) == 1
            and any(t.endswith("...") for t in self.targets)
        )

    def deduplicate_targets(self) -> None:
        targets = self.targets
        if not targets:
            return
        glob_targets = [target for target in targets if target.endswith("/...")]
        non_glob_targets = [target for target in targets if not target.endswith("/...")]
        all_targets = sorted(set(glob_targets)) + sorted(set(non_glob_targets))
        deduplicated_targets = []
        expanded_targets = set()
        for target in all_targets:
            if target.endswith("/...") or target.endswith(":"):
                try:
                    expanded = (
                        subprocess.check_output(["buck2", "query", target])
                        .decode()
                        .strip()
                        .split("\n")
                    )
                    if not all(target in expanded_targets for target in expanded):
                        expanded_targets.update(expanded)
                        deduplicated_targets.append(target)
                except (FileNotFoundError, subprocess.CalledProcessError) as error:
                    LOG.warning("Failed to query target: %s\n%s", target, str(error))
                    deduplicated_targets.append(target)
            elif target not in expanded_targets:
                expanded_targets.add(target)
                deduplicated_targets.append(target)
        deduplicated_targets.sort(key=lambda target: targets.index(target))
        self.targets = deduplicated_targets

    def run_pyre(
        self,
        arguments: List[str],
        description: str,
        should_clean: bool,
        command_input: Optional[str],
        stderr_flag: "subprocess._FILE" = subprocess.PIPE,
    ) -> Optional["subprocess.CompletedProcess[str]"]:
        if should_clean:
            try:
                # If building targets, run clean or space may run out on device!
                LOG.info("Running `buck clean`...")
                subprocess.call(["buck", "clean"], timeout=200)
            except subprocess.TimeoutExpired:
                LOG.warning("Buck timed out. Try running `buck kill` before retrying.")
                return None
            except subprocess.CalledProcessError as error:
                LOG.warning("Error calling `buck clean`: %s", str(error))
                return None
        try:
            LOG.info("%s", description)
            return subprocess.run(
                ["pyre", *arguments],
                stdout=subprocess.PIPE,
                stderr=stderr_flag,
                text=True,
                input=command_input,
            )
        except subprocess.CalledProcessError as error:
            LOG.warning("Error calling pyre: %s", str(error))
            return None

    def get_nested_configuration_paths(self) -> Set[Path]:
        filesystem = get_filesystem()
        nested_configurations = filesystem.list(
            root=self.root, patterns=["**/.pyre_configuration.local"]
        )
        filtered_nested_configurations = {
            path
            for nested_configuration in nested_configurations
            if (path := Path(self.root, nested_configuration)) != self.get_path()
        }
        LOG.info(f"Found {len(filtered_nested_configurations)} nested configurations")
        LOG.debug(f"Nested configurations found: {filtered_nested_configurations}")
        return filtered_nested_configurations

    def get_errors(
        self,
        only_fix_error_code: Optional[int] = None,
        should_clean: bool = True,
        command_input: Optional[str] = None,
        strict: bool = False,
    ) -> Errors:
        local_root_arguments = (
            ["--local-configuration", self.root] if self.is_local else []
        )
        strict_arguments = ["--strict"] if strict else []
        arguments = [*strict_arguments, *local_root_arguments, "--output=json", "check"]
        pyre_output = self.run_pyre(
            arguments=arguments,
            description=f"Checking `{self.root}`...",
            should_clean=self.targets is not None and should_clean,
            command_input=command_input,
        )
        if not pyre_output:
            return Errors.empty()

        stdout = pyre_output.stdout
        if stdout is None:
            return Errors.empty()

        stdout = stdout.strip()

        try:
            errors = Errors.from_json(stdout, only_fix_error_code)
        except UserError as error:
            LOG.info("Error when parsing Pyre error output.")
            LOG.info(f"Pyre stdout: {stdout}\nPyre stderr: {pyre_output.stderr}")
            raise error

        LOG.info("Found %d error%s.", len(errors), "s" if len(errors) != 1 else "")
        return errors



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module provides the logic for the `pyre incremental` command,
which is what is also run when `pyre` is invoked with no subcommand.

What `pyre incremental` does is try to connect to a running daemon
and start one if none is found, then ask for type errors and display
them in the same format as `pyre check` would.
"""

import dataclasses
import enum
import json
import logging
from pathlib import Path
from typing import List, Optional

from .. import (
    backend_arguments,
    command_arguments,
    daemon_socket,
    error,
    frontend_configuration,
    identifiers,
)
from ..language_server import connections
from . import commands, server_event, start

LOG: logging.Logger = logging.getLogger(__name__)

COMMAND_NAME = "incremental"


@dataclasses.dataclass(frozen=True)
class TypeErrors:
    errors: List[error.Error] = dataclasses.field(default_factory=list)
    build_failure: Optional[str] = None


class InvalidServerResponse(Exception):
    pass


class ServerStatus(str, enum.Enum):
    _value_: str
    NEWLY_STARTED = "newly_started_server"
    ALREADY_RUNNING = "already_running_server"

    def __str__(self) -> str:
        return self.value


@dataclasses.dataclass(frozen=True)
class ExitStatus:
    exit_code: commands.ExitCode
    connected_to: ServerStatus


def parse_type_error_response_json(response_json: object) -> TypeErrors:
    try:
        # The response JSON is expected to have one of the following form:
        # `["TypeErrors", [error_json0, error_json1, ...]]` (legacy form)
        # `["TypeErrors", {"errors": [error_json0, ...], "build_failure": "..."}]`
        if (
            isinstance(response_json, list)
            and len(response_json) > 1
            and response_json[0] == "TypeErrors"
        ):
            errors_json = response_json[1]
            if isinstance(errors_json, list):
                return TypeErrors(
                    errors=[
                        error.Error.from_json(error_json) for error_json in errors_json
                    ],
                    build_failure=None,
                )
            elif isinstance(errors_json, dict):
                error_list = errors_json.get("errors", [])
                build_failure = errors_json.get("build_failure", None)
                if isinstance(error_list, list) and (
                    build_failure is None or isinstance(build_failure, str)
                ):
                    return TypeErrors(
                        errors=[
                            error.Error.from_json(error_json)
                            for error_json in error_list
                        ],
                        build_failure=build_failure,
                    )

        raise InvalidServerResponse(
            f"Unexpected JSON response from server: {response_json}"
        )
    except error.ErrorParsingFailure as parsing_error:
        message = f"Unexpected error JSON from server: {parsing_error}"
        raise InvalidServerResponse(message) from parsing_error


def parse_type_error_response(response: str) -> TypeErrors:
    try:
        response_json = json.loads(response)
        return parse_type_error_response_json(response_json)
    except json.JSONDecodeError as decode_error:
        message = f"Cannot parse response as JSON: {decode_error}"
        raise InvalidServerResponse(message) from decode_error


def _read_type_errors(socket_path: Path) -> TypeErrors:
    with connections.connect(socket_path) as (
        input_channel,
        output_channel,
    ):
        # The empty list argument means we want all type errors from the server.
        output_channel.write('["DisplayTypeError", []]\n')
        return parse_type_error_response(input_channel.readline())


def display_type_errors(errors: List[error.Error], output: str) -> None:
    error.print_errors(
        [error.relativize_path(against=Path.cwd()) for error in errors],
        output=output,
    )


def _show_progress_log_and_display_type_errors(
    log_path: Path,
    socket_path: Path,
    output: str,
    remote_logging: Optional[backend_arguments.RemoteLogging],
) -> commands.ExitCode:
    LOG.info("Waiting for server...")
    with start.background_logging(log_path):
        type_errors = _read_type_errors(socket_path)
        display_type_errors(type_errors.errors, output=output)
        if type_errors.build_failure is not None:
            LOG.warning("You may be seeing stale type checking results. Reason:")
            LOG.warning(type_errors.build_failure)
        return (
            commands.ExitCode.SUCCESS
            if len(type_errors.errors) == 0
            else commands.ExitCode.FOUND_ERRORS
        )


def run_incremental(
    configuration: frontend_configuration.Base,
    incremental_arguments: command_arguments.IncrementalArguments,
) -> ExitStatus:
    flavor = identifiers.PyreFlavor.CLASSIC
    socket_path = daemon_socket.get_socket_path(
        configuration.get_project_identifier(),
        flavor=flavor,
    )
    # Need to be consistent with the log symlink location in start command
    log_path = (
        configuration.get_log_directory()
        / flavor.server_log_subdirectory()
        / "server.stderr"
    )
    output = incremental_arguments.output
    remote_logging = backend_arguments.RemoteLogging.create(
        configuration.get_remote_logger(),
        incremental_arguments.start_arguments.get_log_identifier(),
    )
    try:
        exit_code = _show_progress_log_and_display_type_errors(
            log_path, socket_path, output, remote_logging
        )
        return ExitStatus(
            exit_code=exit_code, connected_to=ServerStatus.ALREADY_RUNNING
        )
    except connections.ConnectionFailure:
        pass

    if incremental_arguments.no_start:
        raise commands.ClientException("Cannot find a running Pyre server.")

    LOG.info("Cannot find a running Pyre server. Starting a new one...")
    start_status = start.run(configuration, incremental_arguments.start_arguments)
    if start_status != commands.ExitCode.SUCCESS:
        raise commands.ClientException(
            f"`pyre start` failed with non-zero exit code: {start_status}"
        )
    exit_code = _show_progress_log_and_display_type_errors(
        log_path, socket_path, output, remote_logging
    )
    return ExitStatus(exit_code=exit_code, connected_to=ServerStatus.NEWLY_STARTED)


def run(
    configuration: frontend_configuration.Base,
    incremental_arguments: command_arguments.IncrementalArguments,
) -> ExitStatus:
    try:
        return run_incremental(configuration, incremental_arguments)
    except server_event.ServerStartException as error:
        raise commands.ClientException(
            f"{error}", exit_code=error.kind.to_exit_code()
        ) from error



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
The language_server package contains the functionality powering the Pyre client's
IDE integrations. Handles connection logic and communicating LSP features.
"""

from . import (  # noqa F401
    code_navigation_request,
    connections,
    daemon_connection,
    features,
    protocol,
)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
TODO(T132414938) Add a module-level docstring
"""

import argparse
import json
import logging

from pyre_extensions import override

from ..configuration import Configuration
from ..repository import Repository
from .command import Command

LOG: logging.Logger = logging.getLogger(__name__)


class PysaVersionUpdate(Command):
    def __init__(
        self,
        *,
        repository: Repository,
        hash: str,
        no_commit: bool,
    ) -> None:
        super().__init__(repository)
        self._hash: str = hash
        self._no_commit: bool = no_commit

    @staticmethod
    def from_arguments(
        arguments: argparse.Namespace, repository: Repository
    ) -> "PysaVersionUpdate":
        return PysaVersionUpdate(
            repository=repository,
            hash=arguments.hash,
            no_commit=arguments.no_commit,
        )

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        super(PysaVersionUpdate, PysaVersionUpdate).add_arguments(parser)
        parser.set_defaults(command=cls.from_arguments)
        parser.add_argument("hash", help="Hash of new Pysa version")
        parser.add_argument(
            "--no-commit", action="store_true", help="Keep changes in working state."
        )

    @override
    def run(self) -> None:
        global_configuration = Configuration.find_project_configuration()

        # Update to new pysa version in `.pyre_configuration`
        configuration = Configuration(global_configuration)
        old_version = configuration.pysa_version
        if not old_version:
            LOG.error(
                "Global configuration at %s has no pysa_version field.",
                global_configuration,
            )
            return
        configuration.set_pysa_version(self._hash)
        configuration.write()

        # Update to new pysa version in `.pysa_configuration`
        path = Configuration.find_parent_file(".pysa_configuration")
        if path:
            # TODO(T224086333): move `.pysa_configuration`'s `pysa_version` to less nested `version` field
            contents = json.loads(path.read_text(encoding="utf-8"))
            contents["pyre_configuration"]["pysa_version"] = self._hash
            path.write_text(json.dumps(contents, indent=2), encoding="utf-8")



import sys
from _collections_abc import dict_keys
from collections.abc import Sequence
from typing import Any
from typing_extensions import deprecated

__all__ = ["symtable", "SymbolTable", "Class", "Function", "Symbol"]

if sys.version_info >= (3, 13):
    __all__ += ["SymbolTableType"]

def symtable(code: str, filename: str, compile_type: str) -> SymbolTable: ...

if sys.version_info >= (3, 13):
    from enum import StrEnum

    class SymbolTableType(StrEnum):
        MODULE = "module"
        FUNCTION = "function"
        CLASS = "class"
        ANNOTATION = "annotation"
        TYPE_ALIAS = "type alias"
        TYPE_PARAMETERS = "type parameters"
        TYPE_VARIABLE = "type variable"

class SymbolTable:
    def __init__(self, raw_table: Any, filename: str) -> None: ...
    if sys.version_info >= (3, 13):
        def get_type(self) -> SymbolTableType: ...
    else:
        def get_type(self) -> str: ...

    def get_id(self) -> int: ...
    def get_name(self) -> str: ...
    def get_lineno(self) -> int: ...
    def is_optimized(self) -> bool: ...
    def is_nested(self) -> bool: ...
    def has_children(self) -> bool: ...
    if sys.version_info < (3, 9):
        def has_exec(self) -> bool: ...

    def get_identifiers(self) -> dict_keys[str, int]: ...
    def lookup(self, name: str) -> Symbol: ...
    def get_symbols(self) -> list[Symbol]: ...
    def get_children(self) -> list[SymbolTable]: ...

class Function(SymbolTable):
    def get_parameters(self) -> tuple[str, ...]: ...
    def get_locals(self) -> tuple[str, ...]: ...
    def get_globals(self) -> tuple[str, ...]: ...
    def get_frees(self) -> tuple[str, ...]: ...
    def get_nonlocals(self) -> tuple[str, ...]: ...

class Class(SymbolTable):
    if sys.version_info < (3, 16):
        @deprecated("deprecated in Python 3.14, will be removed in Python 3.16")
        def get_methods(self) -> tuple[str, ...]: ...

class Symbol:
    def __init__(
        self, name: str, flags: int, namespaces: Sequence[SymbolTable] | None = None, *, module_scope: bool = False
    ) -> None: ...
    def is_nonlocal(self) -> bool: ...
    def get_name(self) -> str: ...
    def is_referenced(self) -> bool: ...
    def is_parameter(self) -> bool: ...
    if sys.version_info >= (3, 14):
        def is_type_parameter(self) -> bool: ...

    def is_global(self) -> bool: ...
    def is_declared_global(self) -> bool: ...
    def is_local(self) -> bool: ...
    def is_annotated(self) -> bool: ...
    def is_free(self) -> bool: ...
    if sys.version_info >= (3, 14):
        def is_free_class(self) -> bool: ...

    def is_imported(self) -> bool: ...
    def is_assigned(self) -> bool: ...
    if sys.version_info >= (3, 14):
        def is_comp_iter(self) -> bool: ...
        def is_comp_cell(self) -> bool: ...

    def is_namespace(self) -> bool: ...
    def get_namespaces(self) -> Sequence[SymbolTable]: ...
    def get_namespace(self) -> SymbolTable: ...

class SymbolTableFactory:
    def new(self, table: Any, filename: str) -> SymbolTable: ...
    def __call__(self, table: Any, filename: str) -> SymbolTable: ...



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
strict_default contains a library for running a codemod across a codebase that turns pyre-strict on. Optionally, we can also remove the #pyre-strict and add #pyre-unsafe headers from files where the error count is below a certain threshhold.
"""

from __future__ import annotations

import argparse
import logging
from pathlib import Path
from typing import List, Optional

from pyre_extensions import override

from ....client.find_directories import (
    find_global_and_local_root,
    JSON_CONFIGURATION_FILE,
    LOCAL_CONFIGURATION_FILE,
)
from .. import UserError
from ..configuration import Configuration
from ..filesystem import LocalMode, path_exists, remove_local_mode
from ..repository import Repository
from .command import CommandArguments, ErrorSource, ErrorSuppressingCommand

LOG: logging.Logger = logging.getLogger(__name__)


def _get_configuration_path(local_configuration: Optional[Path]) -> Optional[Path]:
    found_root = find_global_and_local_root(
        Path(".") if local_configuration is None else local_configuration
    )
    if found_root is None:
        return None
    else:
        local_root = found_root.local_root
        if local_root:
            return local_root / LOCAL_CONFIGURATION_FILE
        else:
            return found_root.global_root / JSON_CONFIGURATION_FILE


class StrictDefault(ErrorSuppressingCommand):
    def __init__(
        self,
        command_arguments: CommandArguments,
        *,
        repository: Repository,
        local_configuration: Path,
        remove_strict_headers: bool,
        fixme_threshold: int,
        remove_unsafe_headers: bool,
        remove_strict_flag: bool,
        paths: List[str] | None,
    ) -> None:
        super().__init__(command_arguments, repository)
        self._local_configuration: Path = local_configuration
        self._remove_strict_headers: bool = remove_strict_headers
        self._fixme_threshold: int = fixme_threshold
        self._remove_unsafe_headers: bool = remove_unsafe_headers
        self._remove_strict_flag: bool = remove_strict_flag
        self._paths: List[str] | None = paths

    @staticmethod
    def from_arguments(
        arguments: argparse.Namespace, repository: Repository
    ) -> "StrictDefault":
        command_arguments = CommandArguments.from_arguments(arguments)
        return StrictDefault(
            command_arguments,
            repository=repository,
            local_configuration=arguments.local_configuration,
            remove_strict_headers=arguments.remove_strict_headers,
            fixme_threshold=arguments.fixme_threshold,
            remove_unsafe_headers=arguments.remove_unsafe_headers,
            remove_strict_flag=arguments.remove_strict_flag,
            paths=arguments.paths,
        )

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        super(StrictDefault, cls).add_arguments(parser)
        parser.set_defaults(command=cls.from_arguments)
        parser.add_argument(
            "-l",
            "--local-configuration",
            type=path_exists,
            help="Path to project root with local configuration",
        )
        parser.add_argument(
            "--remove-strict-headers",
            action="store_true",
            help="Delete unnecessary `# pyre-strict` headers.",
        )
        parser.add_argument(
            "--fixme-threshold",
            type=int,
            default=None,
            help="Mark file as unsafe if fixme count exceeds threshold.",
        )
        parser.add_argument(
            "--remove-unsafe-headers",
            action="store_true",
            help="Remove `# pyre-unsafe` headers and replace with `# pyre-fixmes` if the number of new suppressions is under the given fixme threshold",
        )
        parser.add_argument(
            "--remove-strict-flag",
            action="store_true",
            help="Remove the strict flag completely. Useful for repositories where configurations are strict by default.",
        )
        parser.add_argument(
            "--paths",
            default=None,
            nargs="+",
            help="Paths to convert to strict default. Defaults to all paths in the current directory.",
        )

    def _commit_changes(self) -> None:
        title = f"Convert {self._local_configuration} to use strict default"
        summary = (
            "Turning on strict default; files with more than "
            + f"{self._fixme_threshold} errors opted-out of strict."
        )
        self._repository.commit_changes(
            commit=(not self._no_commit),
            title=title,
            summary=summary,
            set_dependencies=False,
        )

    @override
    def run(self) -> None:
        configuration_path = _get_configuration_path(self._local_configuration)
        if configuration_path is None:
            raise UserError("Cannot find a path to configuration")
        configuration = Configuration(configuration_path)
        LOG.info("Processing %s", configuration.get_directory())
        if self._remove_strict_flag:
            configuration.use_strict_default()
        else:
            configuration.add_strict()
        configuration.write()

        argument_paths = self._paths
        if argument_paths is None or len(argument_paths) == 0:
            source_paths = configuration.get_source_paths()
        else:
            LOG.info("Running on specific paths: %s", ", ".join(argument_paths))
            source_paths = (Path(path) for path in argument_paths)

        modes = [
            *([LocalMode.STRICT] if self._remove_strict_headers else []),
            *([LocalMode.UNSAFE] if self._remove_unsafe_headers else []),
        ]

        if self._remove_strict_headers or self._remove_unsafe_headers:
            for path in source_paths:
                remove_local_mode(path, modes)

        self._get_and_suppress_errors(
            configuration,
            error_source=ErrorSource.GENERATE,
            fixme_threshold=self._fixme_threshold,
            fixme_threshold_fallback_mode=LocalMode.UNSAFE,
        )
        self._commit_changes()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
TODO(T132414938) Add a module-level docstring
"""

import argparse
import logging
import subprocess
from pathlib import Path
from typing import Optional

from pyre_extensions import override

from .. import UserError
from ..configuration import Configuration
from ..filesystem import path_exists
from ..repository import Repository
from .command import CommandArguments, ErrorSuppressingCommand
from .consolidate_nested_configurations import consolidate_nested

LOG: logging.Logger = logging.getLogger(__name__)


class FixConfiguration(ErrorSuppressingCommand):
    def __init__(
        self, command_arguments: CommandArguments, *, repository: Repository, path: Path
    ) -> None:
        super().__init__(command_arguments, repository)
        self._path: Path = path
        self._configuration: Optional[Configuration] = Configuration(
            path / ".pyre_configuration.local"
        )

    @staticmethod
    def from_arguments(
        arguments: argparse.Namespace, repository: Repository
    ) -> "FixConfiguration":
        command_arguments = CommandArguments.from_arguments(arguments)
        return FixConfiguration(
            command_arguments, repository=repository, path=arguments.path
        )

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        super(FixConfiguration, cls).add_arguments(parser)
        parser.set_defaults(command=cls.from_arguments)
        parser.add_argument(
            "path",
            help="Path to project root with local configuration",
            type=path_exists,
        )

    def _remove_bad_targets(self) -> None:
        configuration = self._configuration
        if not configuration:
            return
        targets = configuration.targets
        if not targets:
            return
        buildable_targets = []
        for target in targets:
            build_command = ["buck", "query", target]
            try:
                subprocess.check_output(build_command, timeout=30)
            except subprocess.TimeoutExpired:
                buildable_targets.append(target)
            except subprocess.CalledProcessError:
                LOG.info(f"Removing bad target: {target}")
                pass
            else:
                buildable_targets.append(target)
        if len(buildable_targets) == 0 and not configuration.source_directories:
            LOG.info(f"Removing empty configuration at: {configuration.get_path()}")
            self._repository.remove_paths([configuration.get_path()])
            self._configuration = None
        else:
            configuration.targets = buildable_targets
            configuration.write()

    def _consolidate_nested(self) -> None:
        parent_local_configuration_path = Configuration.find_parent_file(
            ".pyre_configuration.local", self._path.parent
        )
        if not parent_local_configuration_path:
            return
        parent_local_configuration = Configuration(parent_local_configuration_path)
        ignored_subdirectories = parent_local_configuration.ignore_all_errors or []
        if (
            str(self._path.relative_to(parent_local_configuration_path.parent))
            in ignored_subdirectories
        ):
            return
        LOG.info(
            f"Consolidating with configuration at: {parent_local_configuration_path}"
        )
        consolidate_nested(
            self._repository,
            parent_local_configuration_path,
            [self._path / ".pyre_configuration.local"],
        )
        self._configuration = parent_local_configuration

    def _commit_changes(self) -> None:
        title = "Fix broken configuration for {}".format(str(self._path))
        self._repository.commit_changes(
            commit=(not self._no_commit),
            title=title,
            summary="Cleaning up broken pyre configurations by removing targets "
            + "that cannot build and removing nested configurations where applicable.",
            reviewers=["pyre", "sentinel"],
        )

    @override
    def run(self) -> None:
        self._remove_bad_targets()
        self._consolidate_nested()

        # Clean any revealed errors.
        configuration = self._configuration
        if configuration:
            try:
                self._get_and_suppress_errors(configuration)
            except UserError as error:
                LOG.warning(
                    f"Configuration at {configuration.get_path()} still "
                    + f"does not build:\n{str(error)}."
                )
                LOG.warning("Discarding changes.")
                self._repository.revert_all(remove_untracked=False)
                return
        self._commit_changes()
